import types
from typing import Callable
from .option_builder import build_option


class Block(object):

    def __init__(self, name: str = 'Block') -> None:
        # Initialise the Block object

        # To set the name of the Block, default = Block
        # Reference to the type of the Block
        self._type = name
        # Title of the block on the editor
        self._name = name

        # To set the defaults for inputs, outputs, options
        self._inputs = {}
        self._outputs = {}
        self._options = {}
        self._interface_names = []

    def __repr__(self) -> str:
        return f'<barfi.Block of type `{self._type}` at {hex(id(self))}>'

    def __str__(self) -> str:        
        inputs_name = [input['name'] for input in self._inputs]
        outputs_name = [output['name'] for output in self._outputs]
        options_name = [option['name'] for option in self._options]
        line_1 = f'barfi.Block of type {self._type} with name {self._name} \n'
        line_2 = f'Inputs: {inputs_name!r} \n'
        line_3 = f'Outputs: {outputs_name!r} \n'
        line_4 = f'Options: {options_name!r} '
        return line_1 + line_2 + line_3 + line_4

    def add_input(self, name: str = None, value = None) -> None:
        """
        A function defined to add an Input interface to the Block

        Use as:
            self.addInput(name='Input Name')

        Interface options:
            name (str)  : The name of the Input interface.
            value (any)  : The default value for this input interface.
        """        
        if name:
            if name in self._interface_names: raise ValueError(f'name: {name} already exists as an interface to the Block.')
            self._inputs[name] = {'value': value, 'id': None}
            self._interface_names.append(name)
        else:
            in_nos = len(self._inputs)
            name = 'Input ' + str(in_nos + 1)
            self._inputs[name] = {'value': value, 'id': None}  
            self._interface_names.append(name)      

    def add_output(self, name: str = None, value = None) -> None:
        """
        A function defined to add an Output interface to the Block

        Use as:
            self.addOutput(name='Output Name')

        Interface options:
            name (str)  : The name of the Output interface.
            value (any)  : The default value for this output interface.
        """
        if name:
            if name in self._interface_names: raise ValueError(f'name: {name} already exists as an interface to the Block.')
            self._outputs[name] = {'value': value, 'id': None}
            self._interface_names.append(name)      
        else:
            out_nos = len(self._outputs)
            name = 'Output ' + str(out_nos + 1)
            self._outputs[name] = {'value': value, 'id': None}
            self._interface_names.append(name)

    def get_interface(self, name: str):
        
        if name in self._inputs:
            return self._inputs[name]['value']  
        elif name in self._outputs:
            return self._outputs[name]['value']    
        else:
            raise ValueError(f'No interface with name: {name} found for Block')   
            return None

    def set_interface(self, name: str, value) -> None:
        if name in self._inputs:
            self._inputs[name]['value'] = value            
        elif name in self._outputs:
            self._outputs[name]['value'] = value
        else:
            raise ValueError(f'No interface with name: {name} found for Block')   

    def _set_interface_id(self, name: str, id: str) -> None:
        if name in self._inputs:
            self._inputs[name]['id'] = id
        elif name in self._outputs:            
            self._outputs[name]['id'] = id
        else:
            raise ValueError(f'No interface with name: {name} found for Block')   

    def add_option(self, name: str, type: str, **kwargs) -> None:
        """
        A function defined to add interactive Option interface to the Block

        Use as:
            self.addOption(name=name, type=type, **kwargs)

        Interactive options interface:
            name (str)  : The name of the Option interface.
            type (str)  : The type of the Option interface. 'checkbox', 'input', 'integer', 'number', 'select', 'slider', 'text'.
            value       : The default value for the option. Depends on the option chosen.

            Additional properties depending on the type of Option interface.

        """

        assert isinstance(
            name, str), "Error: 'name' argument should be of type string."
        assert isinstance(
            type, str), "Error: 'type' argument should be of type string."
        assert type in ['checkbox', 'input', 'integer', 'number', 'select', 'slider',
                        'display'], 'Error: Option "type" is not of standard Option interface parameter.'

        if name in self._options:
            raise ValueError(f'Option with name: {name} aready exists in Block.')

        _option = build_option(name, type, kwargs)

        self._options[_option['name']] = _option

    def set_option(self, name: str, **kwargs): 
        # Can only set the 'value' property for now. 
        if name in self._options:
            for arg, value in kwargs.items():
                if arg in self._options[name]:
                    if arg in ['value']:
                        self._options[name][arg] = value
                    else:
                        raise ValueError(f'Cannot set or invalid property: {arg} for Block option.')
                else:
                    raise ValueError(f'Property: {arg} is not a valid option property for {name}.')
        else: 
            raise ValueError(f'Option name: {name} does not exist in Block.')

    def get_option(self, name: str): 
        if name in self._options:
            return self._options[name]['value']
        else: 
            raise ValueError(f'Option name: {name} does not exist in Block.')

    def _export(self):
        _inputs_export = []
        _outputs_export = []
        _options_export = []
        for key, _ in self._inputs.items():
            _inputs_export.append({'name': key})
        for key, _ in self._outputs.items():
            _outputs_export.append({'name': key})
        for _, item in self._options.items():
            _options_export.append(item)

        return {'name': self._name, 'inputs': _inputs_export, 'outputs': _outputs_export, 'options': _options_export}

    def _on_compute():
        pass

    def add_compute(self, _func: Callable):
        self._on_compute = types.MethodType(_func, self)



import networkx as nx
import copy
from .block_builder import Block
from typing import List, Dict


class ComputeEngine(object):
    def __init__(self, blocks: List[Block] = [], **kwargs) -> None:
        # Initialise the Compute Enginge
        # The primitive block types that are tied to this compute engine
        self._blocks = blocks

        # The state of the editor that is stored as JSON/Dict that represents
        # the active blocks and links on the editor
        self._editor_state = {}

        # Active blocks extracted from the editor_state.
        self._active_blocks = {}

        # Mapping all the active blocks and connections to their id and names
        self._map_block_id_name = {}
        self._map_interface_id_block_id = {}
        self._map_interface_id_name = {}
        self._map_link_interface_id_from_to = {}
        self._map_link_interface_id_to_from = {}

        # Create a directed-acyclic-graph of the blocks and links
        # The computational graph
        self._graph = nx.DiGraph()

        # Result from the execution for the compute engine
        self._result = {}

    def add_editor_state(self, editor_state):
        self._editor_state = editor_state

    def get_result(self) -> Dict:
        # return self._active_blocks
        return self._result

    def _map_block_link(self):
        if bool(self._editor_state):

            for _block in self._editor_state['nodes']:

                # Create a child block object for the active the block and associate with its id
                _parent_block = next(
                    _b for _b in self._blocks if _b._type == _block['type'])
                # Create an independent deep copy of the parent block type
                _child_block = copy.deepcopy(_parent_block)
                _child_block._name = _block['name']

                # Map the active block id to its name
                self._map_block_id_name[_block['id']] = _block['name']

                # Map the active interfaces/Links to the active block
                _block_interfaces = {}
                for _interface in _block['interfaces']:
                    self._map_interface_id_block_id[_interface[1]
                                                    ['id']] = _block['id']
                    self._map_interface_id_name[_interface[1]
                                                ['id']] = _interface[0]
                    _child_block._set_interface_id(
                        name=_interface[0], id=_interface[1]['id'])
                    _block_interfaces[_interface[0]] = _interface[1]                

                for _option in _block['options']:                    
                    _child_block.set_option(name=_option[0], value=_option[1])                    

                # Active blocks build from the base-blocks and editor-state
                self._active_blocks[_block['id']] = {
                    'block': _child_block, 'interfaces': _block_interfaces, 'type': _block['type'], 'name': _block['name']}

                # Add block to the DAG as a node
                self._graph.add_node(_block['id'], name=_block['name'])

            for _connection in self._editor_state['connections']:
                from_node = self._map_interface_id_block_id[_connection['from']]
                to_node = self._map_interface_id_block_id[_connection['to']]
                if _connection['from'] not in self._map_link_interface_id_from_to:
                    self._map_link_interface_id_from_to[_connection['from']] = [
                    ]
                self._map_link_interface_id_from_to[_connection['from']].append(
                    _connection['to'])
                self._map_link_interface_id_to_from[_connection['to']
                                                    ] = _connection['from']

                # Add the _connection to the DAG as an edge
                self._graph.add_edge(from_node, to_node,
                                     edge_id=_connection['id'])

            if not nx.is_directed_acyclic_graph(self._graph):
                raise('Cycle(s) detected. Not supported by `barfi` at the moment.')
            else:
                _compu_order = [self._map_block_id_name[node]
                                for node in nx.topological_sort(self._graph)]

            # TODO morph _result to _active_blocks and have only one of them.
            # TODO add the interface info to the block
            for block_id, block in self._active_blocks.items():
                self._result[block['name']] = {'block': block['block'],
                                               'type': block['type'],
                                               'interfaces': {}}
                for link_id, link in block['interfaces'].items():
                    self._result[block['name']]['interfaces'][link_id] = {}
                    _interface_id = link['id']

                    if _interface_id in self._map_link_interface_id_from_to:
                        self._result[block['name']
                                     ]['interfaces'][link_id]['type'] = 'output'
                        self._result[block['name']
                                     ]['interfaces'][link_id]['to'] = {}
                        for to_id in self._map_link_interface_id_from_to[_interface_id]:
                            to_name = self._map_interface_id_name[to_id]
                            to_block_id = self._map_interface_id_block_id[to_id]
                            to_block_name = self._map_block_id_name[to_block_id]
                            self._result[block['name']
                                         ]['interfaces'][link_id]['to'][to_block_name] = to_name

                    if _interface_id in self._map_link_interface_id_to_from:
                        self._result[block['name']
                                     ]['interfaces'][link_id]['type'] = 'intput'
                        self._result[block['name']
                                     ]['interfaces'][link_id]['from'] = {}
                        from_id = self._map_link_interface_id_to_from[_interface_id]
                        from_name = self._map_interface_id_name[from_id]
                        from_block_id = self._map_interface_id_block_id[from_id]
                        from_block_name = self._map_block_id_name[from_block_id]
                        self._result[block['name']
                                     ]['interfaces'][link_id]['from'][from_block_name] = from_name

    def _execute_compute(self):
        if bool(self._editor_state):

            for node in nx.topological_sort(self._graph):
                self._active_blocks[node]['block']._on_compute()
                for key, value in self._active_blocks[node]['block']._outputs.items():
                    try:
                        for find_to in self._map_link_interface_id_from_to[value['id']]:
                            find_to_block = self._map_interface_id_block_id[find_to]
                            self._active_blocks[find_to_block]['block'].set_interface(
                                name=self._map_interface_id_name[find_to], value=value['value'])
                    except:
                        pass
